# 'around every corner' experiment - formerly 'paranoia'
# by Adam Gashlin (hcs)

import warnings
warnings.filterwarnings(category=RuntimeWarning,action='ignore')
import sys
import math
from numpy import argwhere
from numpy import flatnonzero
import pygame.display
import pygame.event
import pygame.surfarray
#import pygame.image
import level
import fov
import digits

Block = level.Block
clip_world = level.clip_world
fov_polygon = fov.fov_polygon

pygame.init

clock = pygame.time.Clock()

# ***** Graphics setup *****

# set up window
resolution = (240,160)
square_res = max(resolution[0],resolution[1])

# each engine pixel scaled up to 3x3 real pixels
real_resolution = (resolution[0]*3, resolution[1]*3)

main_win = pygame.display.set_mode( real_resolution )
pygame.display.set_caption('around every corner')

# full_win is the final 8-bit texture, it is the only one with our real colors 
BLACK = 0
WHITE = 1

palette = [pygame.Color('black'), pygame.Color('white')]
full_win = pygame.Surface( real_resolution, 0, 8 )
full_win.set_palette(palette)

# small_win is the last stage of engine-resolution
small_win = pygame.Surface( resolution, 0, 8 )

# arena is the fade-processed historical FOV
arena = pygame.Surface( resolution, 0, 8 )
# old_arena was the arena as of the end of the last cycle
old_arena = arena.copy()

# used to draw the creeping pixels
creep_mask = pygame.Surface(resolution, 0, 8)
creep_mask.set_colorkey(2)

# intermediate surfaces for rendering current FOV (aka eyebeam)
eyebeam = pygame.Surface(resolution, 0, 8)
eyebeam.set_colorkey(2)
eyebeam_mask = pygame.Surface(resolution, 0, 8)
eyebeam_mask.set_colorkey(0)

# ***** Mechanics *****

guy_speed = 2
fade_speed = 1 # how many cycles of creep to process each frame

FPS = 30

fov_angle = math.pi*3/8 # 67.5 degrees, total field is 135 degrees

# what to do when direction keys are pressed (velocity)
dir_lookup = {pygame.K_UP   : (0, -guy_speed),
              pygame.K_DOWN : (0, guy_speed),
              pygame.K_LEFT : (-guy_speed, 0),
              pygame.K_RIGHT: (guy_speed, 0)}

# ***** Level layout *****

#level.dump_level("lev1.txt", world)
world = level.load_level("lev1.txt")

def scroll(dist):
    global world_pos, guy_pos, blocks
    world_pos = (world_pos[0]+dist[0], world_pos[1]+dist[1])
    guy_pos = (guy_pos[0]-dist[0], guy_pos[1]-dist[1])
    blocks = clip_world(world, world_pos, resolution)
    old_arena.scroll(-dist[0],-dist[1])
    arena.scroll(-dist[0],-dist[1])

    left_mask_len = 0
    right_mask_len = 0
    if dist[0] > 0: # left scroll
        left_mask_len = 1
        right_mask_len = dist[0]
    if dist[0] < 0: # right scroll
        left_mask_len = -dist[0]
        right_mask_len = 1

    if left_mask_len != 0:
        arena.fill(BLACK, pygame.Rect((0,0),(left_mask_len,resolution[1])))
        old_arena.fill(BLACK, pygame.Rect((0,0),(left_mask_len,resolution[1])))
        pygame.draw.line(old_arena, WHITE, (left_mask_len-1,0),(left_mask_len-1,resolution[1]-1))
    if right_mask_len != 0:
        arena.fill(BLACK, pygame.Rect((resolution[0]-right_mask_len,0),(right_mask_len, resolution[1])))
        old_arena.fill(BLACK, pygame.Rect((resolution[0]-right_mask_len,0),(right_mask_len, resolution[1])))
        pygame.draw.line(old_arena, WHITE, (resolution[0]-right_mask_len,0),(resolution[0]-right_mask_len,resolution[1]-1))

def creep_eyebeam():
    global eyebeam, eyebeam_mask, eyebeam_poly, old_arena
    # say that the outline was lit before the creep that just happened
    pygame.draw.polygon(old_arena, WHITE, eyebeam_poly, 1)

    # We must now remove the edge of the eyebeam, to begin the creep.
    pygame.draw.polygon(eyebeam, 2, eyebeam_poly, 1)

    # eyebeam_mask is additional pixels on bottom fringes of the polygon
    # that are not removed by drawing the edge above
    eyebeam_flipped = []

    for p in eyebeam_poly:
        eyebeam_flipped.append((p[0],resolution[1]-p[1]-1))

    eyebeam_mask.fill(0)
    # Draw polygon normally
    pygame.draw.polygon(eyebeam_mask, 2, eyebeam_poly, 0)
    # Flip and overwrite with flipped polygon. (0 is colorkey for
    #   eyebeam_mask). This will leave as the only color '2' pixels those
    #   that were part of the original but not the flipped. draw.polygon
    #   fills past the end of descending lines, this will expose those
    #   since they are ascending lines when flipped.
    eyebeam_mask = pygame.transform.flip(eyebeam_mask,0,1)
    pygame.draw.polygon(eyebeam_mask, 0, eyebeam_flipped, 0)
    # Flip back into original orientation
    eyebeam_mask = pygame.transform.flip(eyebeam_mask,0,1)
    # Blit more '2's (colorkey) onto the original eyebeam to completely
    # remove the edge.
    eyebeam.blit(eyebeam_mask, (0,0))

# ***** Per-game loop
while 1:
    # ***** Per-game Inits
    reset_delay = 0

    frame = 0
    score = 0

    world_pos = (0,0)
    guy_pos = (40,80)
    guy_pos = (100,80)
    guy_vel = (0,0)
    guy_angle = 0
    eye_open = True
    paused = False

    # how many direction keys are currently held (init to zero lets us ignore
    # anything held when we start as long as another isn't pressed first)
    direction_count = 0

    # initially set up clipped blocks
    scroll((0,0))

    # initial lighting (eyebeam in all directions)
    old_arena.fill(BLACK)
    arena.fill(BLACK)
    for angle in [0,90,180,-90]:
        eyebeam_poly = fov_polygon(resolution, guy_pos, fov_angle, angle, blocks)
        eyebeam.fill(2) # 2 is eyebeam's colorkey
        pygame.draw.polygon(eyebeam, WHITE, eyebeam_poly, 0)
        pygame.draw.polygon(eyebeam, WHITE, eyebeam_poly, 1)
        creep_eyebeam()
        arena.blit(eyebeam, (0,0))
    # only creeped point common to all 
    arena.set_at(guy_pos, WHITE)

    # ***** Main Loop *****

    while 1:
        # event-driven stuff
        world_scroll = (0,0)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()
            if event.type == pygame.KEYDOWN:
                if event.key in (pygame.K_ESCAPE, pygame.K_q):
                    sys.exit()
                if event.key in dir_lookup:
                    guy_vel = (guy_vel[0]+dir_lookup[event.key][0],guy_vel[1]+dir_lookup[event.key][1])
                    direction_count += 1
                if event.key == pygame.K_SPACE:
                    paused = not paused
            if event.type == pygame.KEYUP:
                if event.key in dir_lookup and direction_count != 0:
                    guy_vel = (guy_vel[0]-dir_lookup[event.key][0],guy_vel[1]-dir_lookup[event.key][1])
                    direction_count -= 1

        if paused:
            continue

        # move!
        world_scroll = (guy_vel[0],0)
        if world_scroll != (0,0):
            scroll(world_scroll)

        # face generally in the direction of motion, fixed to cardinal
        # directions, preferring horizontal
        #if guy_vel[0] > 0:
        #    guy_angle = 0
        #elif guy_vel[0] < 0:
        #    guy_angle = 180
        #elif guy_vel[1] < 0:
        #    guy_angle = 90
        #elif guy_vel[1] > 0:
        #    guy_angle = -90

        #guy_pos = (guy_pos[0],guy_pos[1]+1)
        #if guy_pos[1] == resolution[1]-1:
        #    guy_pos = (guy_pos[0]+1,1)
        #if guy_pos[0] == resolution[0]-1:
        #    guy_pos = (1,1)

        # ***** Render *****
        small_win.blit(arena, (0,0))

        guy_old_pos = guy_pos
        guy_pos = (guy_pos[0] + guy_vel[0], guy_pos[1] + guy_vel[1])
        if guy_pos[1] < 0:
            guy_pos = (guy_pos[0],0)
        if guy_pos[1] > resolution[1]-1:
            guy_pos = (guy_pos[0],resolution[1]-1)
 
        # collision detection
        if small_win.map_rgb(small_win.get_at(guy_pos)) != WHITE:
            eye_open = False

        # step along the path (diagonal until horiz or vert, then straight) and
        # ensure that we could have walked it
        while guy_old_pos != guy_pos:
            if small_win.map_rgb(small_win.get_at(guy_old_pos)) != WHITE:
                eye_open = False

            if guy_old_pos[0] < guy_pos[0]:
                guy_old_pos = (guy_old_pos[0]+1, guy_old_pos[1])
            elif guy_old_pos[0] > guy_pos[0]:
                guy_old_pos = (guy_old_pos[0]-1, guy_old_pos[1])
            if guy_old_pos[1] < guy_pos[1]:
                guy_old_pos = (guy_old_pos[0], guy_old_pos[1]+1)
            elif guy_old_pos[1] > guy_pos[1]:
                guy_old_pos = (guy_old_pos[0], guy_old_pos[1]-1)

        # god mode (debug)
        #eye_open = True

        # generate eyebeam
        eyebeam_poly = []
        if eye_open:
            eyebeam_poly = fov_polygon(resolution, guy_pos, fov_angle, guy_angle, blocks)

            # render beam
            eyebeam.fill(2) # 2 is eyebeam's colorkey
            pygame.draw.polygon(eyebeam, WHITE, eyebeam_poly, 0)
            pygame.draw.polygon(eyebeam, WHITE, eyebeam_poly, 1)

            # render eyebeam directly
            #small_win.blit(eyebeam, (0,0))

        # Update creep
        for i in range(fade_speed):
            arena_arr = pygame.surfarray.array2d(arena)
            old_arena_arr = pygame.surfarray.pixels2d(old_arena)
            # If a pixel is newly dark, it creeps
            creep_mask.fill(2)
            creep_mask_arr = pygame.surfarray.pixels2d(creep_mask)
            for (x,y) in argwhere(arena_arr < old_arena_arr):
                creep_mask_arr[x,y] = 0
            del creep_mask_arr
            arena.blit(creep_mask, (-1,0))
            arena.blit(creep_mask, (1,0))
            arena.blit(creep_mask, (0,-1))
            arena.blit(creep_mask, (0,1))

            old_arena = pygame.surfarray.make_surface(arena_arr)

        if eye_open:
            creep_eyebeam()
            arena.blit(eyebeam, (0,0))

            score += 1
        else:
            # death reset timer
            arena_arr = pygame.surfarray.array2d(arena)
            live_count = len(flatnonzero(arena_arr))
            if live_count == 0:
                reset_delay += 1
            # delay for 1 second
            if reset_delay == FPS:
                print "Score:",score
                break

        # avatar
        small_win.set_at(guy_pos, pygame.Color('red'))

        # show block outlines (debug)
        #for b in blocks:
        #    pygame.draw.lines(small_win, pygame.Color('green'), 1, b.points)

        # scale up and blit to real window
        # Doesn't scale directly to main_win because main_win isn't 8-bit
        small_win.blit(digits.render_number(score), (10,10))
        pygame.transform.scale(small_win, real_resolution, full_win)
        main_win.blit(full_win, (0,0))

        pygame.display.update()

        frame += 1

        #print pygame.time.get_ticks() - start_time
        #print gc.get_count()

        clock.tick(FPS)
        #print "frame",frame

        # frame dumper (debug)
        #pygame.image.save(full_win, "frame{0:0>4d}.bmp".format(frame))

# "for when the darkness finds you, the sun will cease to shine"
#  - Jonathan Coulton, "Brookline"
