# computation of a polygonal Field of View

import math

def to_polar2(point, origin):
    deltax = point[0]-origin[0]
    deltay = point[1]-origin[1]
    return (math.atan2(deltay,deltax),math.hypot(deltax,deltay))

def index_blocks(screen_resolution, rotation, screen_origin, blocks, horizontal_first):
    orig_origin = rotation(screen_origin)

    def to_polar(point):
        return to_polar2(point, origin)
    def flip(point):
        return (point[0], screen_resolution[1]-point[1]-1)

    # generate list of points and line segments in the quadrant of interest
    # TODO: would be good to have a quick way to lookup points by lines and vv
    all_points = ([],[]) # will be ((x,y),(theta,r),convex,block_id)
    all_hlines = ([],[]) # will be ((minx,maxx),deltay,block_id)
    all_vlines = ([],[]) # will be (deltax,(miny,maxy),block_id)
    block_no = 0
    for b in blocks:
        count = 0
        horizontal = horizontal_first
        raw_points = map(rotation,b.points)
        for orig_p in raw_points:
            orig_next_p = raw_points[(count+1)%len(raw_points)]
            for (points,hlines,vlines,origin,p,next_p) \
                in [(all_points[0], all_hlines[0], all_vlines[0], orig_origin, orig_p, orig_next_p), \
                    (all_points[1], all_hlines[1], all_vlines[1], flip(orig_origin), flip(orig_p), flip(orig_next_p))]:

                line_in_range = 0
                if p[0] > origin[0] and p[1] <= origin[1]:
                    line_in_range = 1
                    points.append((p,         # cartesian coordinates for point
                                  to_polar(p),# polar relative to origin
                                  b.convex[count],
                                  block_no))
                if not line_in_range and next_p[0] > origin[0] and next_p[1] <= origin[1]:
                    line_in_range = 1
                if line_in_range:
                    if horizontal:
                        hlines.append(
                            ((min(p[0],next_p[0]),max(p[0],next_p[0])),
                             p[1]-origin[1],block_no))
                    else:
                        vlines.append(
                            (p[0]-origin[0],
                             (min(p[1],next_p[1]),max(p[1],next_p[1])),block_no))
            horizontal = not horizontal
            count += 1
        block_no += 1

    return [(all_points[0],all_hlines[0],all_vlines[0]),
            (all_points[1],all_hlines[1],all_vlines[1])]

def trace_fov(resolution, origin, view_angle, points, hlines, vlines):
    def to_polar(point):
        return to_polar2(point, origin)

    # begin drawing
    result = [(int(origin[0]),int(origin[1]))]
    current_xy = origin
    current_theta = -view_angle
    current_r = 0

    # main loop

    have_line_segment = 0
    while 1:
        if len(result) > 1000:
            raise RuntimeError('too many segments, probably stuck')
        #print "len",len(result),"current_xy",current_xy,"current_polar",(current_theta,current_r)

        if not have_line_segment:
            # 1. Find nearest intersecting line segment
            tan_theta = math.tan(current_theta)
            nearest_line = ()
            nearest_dist_squared = max(resolution[0],resolution[1])*10
            nearest_dist_squared *= nearest_dist_squared
            nearest_intersection = ()
            nearest_is_h = True
            for L in hlines:
                deltax = round(L[1] / tan_theta,1)  # deltay/tan
                x = origin[0] + deltax
                if deltax > 0 and x >= L[0][0] and x <= L[0][1]:    # between min and max x
                    dist = deltax*deltax + L[1]*L[1] 
                    #print L, dist
                    if dist < nearest_dist_squared:
                        #print "yes!"
                        intersection = (x,origin[1]+L[1])
                        nearest_dist_squared = dist
                        nearest_line = L
                        nearest_is_h = True
                        nearest_intersection = intersection
            for L in vlines:
                deltay = round(L[0] * tan_theta,1)  # deltax*tan
                y = origin[1] + deltay
                if deltay < 0 and y >= L[1][0] and y <= L[1][1]:    # between min and max y
                    dist = L[0]*L[0] + deltay*deltay
                    #print L, dist
                    if dist < nearest_dist_squared:
                        #print "yes!"
                        intersection = (origin[0]+L[0],y)
                        nearest_dist_squared = dist
                        nearest_line = L
                        nearest_is_h = False
                        nearest_intersection = intersection

        if nearest_line == ():
            raise RuntimeError("didn\'t find an intersection")

        line_end_xy = (nearest_line[0][1], origin[1]+nearest_line[1]) if \
            nearest_is_h \
            else (origin[0]+nearest_line[0], nearest_line[1][1])
        (line_end_theta, line_end_r) = to_polar(line_end_xy)

        line_other_end_xy = (nearest_line[0][0], origin[1]+nearest_line[1]) if \
            nearest_is_h \
            else (origin[0]+nearest_line[0], nearest_line[1][0])

        # if we land directly on a corner adjust theta down a bit to try and
        # skip past it
        #print nearest_intersection, line_end_xy, line_other_end_xy
        if not have_line_segment:
            if nearest_intersection == line_end_xy or nearest_intersection == line_other_end_xy:
                #print "adjust"
                current_theta += math.pi/4/max(resolution[0],resolution[1])
                if current_theta >= 0: return result
                continue

        # draw a line to intersection
        if result[-1] != nearest_intersection:
            result.append((int(nearest_intersection[0]),int(nearest_intersection[1])))

        # intersection becomes current point
        current_xy = nearest_intersection
        (current_theta, current_r) = to_polar(current_xy)

        # 2. Find the corner with the lowest angle in the triangle formed by the
        # origin, the current point, and the end of the line segment
        lowest_angle = 0
        lowest_angle_point = ()
        for p in points:
            within_segment = 0
            if ((p[1][0] > current_theta and p[1][0] < line_end_theta) or \
                (p[1][0] == current_theta and p[1][0] < current_r) or \
                (p[1][0] == line_end_theta and p[1][0] < line_end_r)) and \
              (    (nearest_is_h and p[0][1] > current_xy[1]) or \
               (not nearest_is_h and p[0][0] < current_xy[0])) and \
              p[1][0] < lowest_angle:
                        lowest_angle = p[1][0]
                        lowest_angle_point = p

        if lowest_angle_point == ():
            if line_end_theta >= 0:
                if result[-1] != (current_xy[0], origin[1]):
                    result.append((int(current_xy[0]), int(origin[1])))
                return result

            # 2b. if there is no qualifying point
            # draw line from current point to endpoint
            if result[-1] != line_end_xy:
                result.append((int(line_end_xy[0]),int(line_end_xy[1])))
            # endpoint becomes current point
            current_xy = line_end_xy
            (current_theta, current_r) = (line_end_theta, line_end_r)

            # find corresponding point
            current_point = ()
            for p in points:
                # match block_id and coordinates
                if p[3] == nearest_line[2] and \
                   p[0] == line_end_xy:
                   if current_point != ():
                       raise RuntimeError('multiple candidate points')
                   current_point = p
            if current_point == ():
                raise RuntimeError("didn't find point for segment endpoint")

            if nearest_is_h:    # 2b.i) if line segment is horizontal
                if current_point[2]:    # 2b.i.I) if corner is convex
                    # continue with Step 1
                    have_line_segment = 0
                    continue
                else:                   # 2b.i.II) if corner is concave
                    # next segment is the one around this corner, vertical
                    nearest_line = ()
                    deltax = current_xy[0] - origin[0]
                    for L in vlines:
                        # match block_id and coordinates
                        if L[2] == current_point[3] and \
                           L[0] == deltax and L[1][0] == current_point[0][1]:
                           if nearest_line != ():
                               raise RuntimeError('multiple candidate lines')
                           nearest_line = L
                           nearest_is_h = 0
                           nearest_intersection = current_xy
                    if nearest_line == ():
                        raise RuntimeError("didn't find line around the corner (concave)")
                    have_line_segment = 1
                    continue
            else:   # 2b.ii) if line segment is vertical
                # 2b.ii.I) corner should always be convex
                if not current_point[2]:
                    raise RuntimeError('expected a convex corner (2b.ii)')
                # next segment is the one around this corner, horizontal
                nearest_line = ()
                deltay = current_xy[1] - origin[1]
                for L in hlines:
                    # match block_id and coordinates
                    if L[2] == current_point[3] and \
                       L[0][0] == current_point[0][0] and L[1] == deltay:
                       if nearest_line != ():
                           raise RuntimeError('multiple candidate lines')
                       nearest_line = L
                       nearest_is_h = 1
                       nearest_intersection = current_xy
                if nearest_line == ():
                    raise RuntimeError("didn't find line around the corner (convex)")
                have_line_segment = 1
                continue

        else:
            # 2c. if there is such a point
            if not lowest_angle_point[2]:
                raise RuntimeError('expected a convex corner (2c)')
            # draw a line to the equal angle on the current line
            tan_lowest_angle = math.tan(lowest_angle)
            intersection = \
                (int(origin[0] + nearest_line[1] / tan_lowest_angle), int(origin[1] + nearest_line[1])) if nearest_is_h else \
                (int(origin[0] + nearest_line[0]), int(origin[1] + nearest_line[0] * tan_lowest_angle))
            if result[-1] != intersection:
                result.append(intersection)
            # draw a line to the intercepting point
            if result[-1] != lowest_angle_point[0]:
                result.append((int(lowest_angle_point[0][0]),int(lowest_angle_point[0][1])))
            # that point becomes the current point
            current_xy = lowest_angle_point[0]
            (current_theta, current_r) = lowest_angle_point[1]

            # find corresponding vertical line segment
            # (if we intercept a corner, the horizontal face is out of sight)
            deltax = current_xy[0] - origin[0]
            nearest_line = ()
            for L in vlines:
                # match block_id and coordinates
                if L[2] == lowest_angle_point[3] and \
                   L[0] == deltax and L[1][0] == current_xy[1]:
                   if nearest_line != ():
                       raise RuntimeError('multiple candidate lines')
                   nearest_line = L
                   nearest_is_h = False
                   nearest_intersection = current_xy
            if nearest_line == ():
                raise RuntimeError("didn't find intercepting segment")

            have_line_segment = 1
            continue
        break

def fov_polygon(resolution, origin, view_angle, facing_angle, blocks):
    # counter-clockwise rotation in screenspace
    rotate0 = lambda p: p
    rotate90 = lambda (x,y): (y,resolution[0]-1-x)
    rotate180 = lambda (x,y): (resolution[0]-1-x,resolution[1]-1-y)
    rotate270 = lambda (x,y): (resolution[1]-1-y,x)
    unrotate90 = lambda (x,y): (y,resolution[1]-1-x)
    unrotate270 = lambda (x,y): (resolution[0]-1-y,x)
    if facing_angle == 0:       # right
        indexing_resolution = resolution
        rotation = rotate0
        unrotation = rotate0
        horizontal_first = True
    elif facing_angle == 90:    # up
        indexing_resolution = (resolution[1],resolution[0])
        rotation = rotate270
        unrotation = unrotate90
        horizontal_first = False
    elif facing_angle == 180:   # left
        indexing_resolution = resolution
        rotation = rotate180
        unrotation = rotate180
        horizontal_first = True
    elif facing_angle == -90:   # down
        indexing_resolution = (resolution[1],resolution[0])
        rotation = rotate90
        unrotation = unrotate270
        horizontal_first = False
    geometry = index_blocks(indexing_resolution, rotation, origin, blocks, horizontal_first)
    rotated_origin = rotation(origin)
    try:
        vertices = trace_fov(resolution, rotated_origin, view_angle,\
                geometry[0][0], geometry[0][1], geometry[0][2])
        backwards_vertices = trace_fov(resolution,\
                (rotated_origin[0], indexing_resolution[1]-rotated_origin[1]-1), view_angle,\
                geometry[1][0], geometry[1][1], geometry[1][2])
    except RuntimeError, e:
        print "Error with",origin,":",e
        return [origin,origin,origin]  # dummy

    backwards_vertices.reverse()
    for p in backwards_vertices[0:-1]:
        vertices.append((p[0],indexing_resolution[1]-p[1]-1))

    return map(unrotation, vertices)

